// Author: Hugo Pointcheval
// Email: git@pcl.ovh
// -----
// File: app.dart
// Created Date: 19/08/2022 12:05:38
// Last Modified: Wed Nov 09 2022
// -----
// Copyright (c) 2022

import 'dart:async';

import 'package:example_router/core/dependency_injection/get_it.dart';
import 'package:example_router/core/routes/router.dart';
import 'package:example_router/core/utils/forms.dart';
import 'package:example_router/presentation/features/home/home_page.dart';
import 'package:example_router/presentation/features/welcome/welcome_page.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:go_router/go_router.dart';
import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

class App extends StatelessWidget {
  final AuthenticationRepository<int> authenticationRepository =
      AuthenticationRepositoryImpl(getIt<AuthenticationLocalDataSource<int>>(),
          getIt<AuthenticationRemoteDataSource>(), (account) async {
    debugPrint('onSignUpSuccess: $account');
    return const Ok(null);
  }, (account) async {
    debugPrint('onAccountChanges: $account');
    return const Ok(null);
  });

  App({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    AuthenticationState? previous;

    final AuthenticationCubit<int> authenticationCubit =
        AuthenticationCubit(authenticationRepository: authenticationRepository);

    final GoRouter router = GoRouter(
      initialLocation: '/',
      routes: AppRouter.routes,
      debugLogDiagnostics: true,
      errorBuilder: (_, __) => const ColoredBox(
        color: Colors.red,
      ),
      refreshListenable: GoRouterRefreshStream(authenticationCubit.stream),
      redirect: (context, state) {
        final authState = authenticationCubit.state;

        if (authState != previous) {
          previous = authState;
          // Check if current user is logged in
          final loggedIn =
              authState.status == AuthenticationStatus.authenticated;

          // Checking if current path is onboarding or not
          final isOnboarding = AppRouter.publicRoutes.contains(state.subloc);

          if (!loggedIn) {
            debugPrint('Not logged');
            if (isOnboarding) {
              return null;
            } else {
              return state.namedLocation(WelcomePage.pageName);
            }
          } else {
            debugPrint('Logged');
            if (isOnboarding) {
              return state.namedLocation(HomePage.pageName);
            } else {
              return null;
            }
          }
        }
        return null;
      },
    );

    return MultiRepositoryProvider(
      providers: [
        RepositoryProvider<AuthenticationRepository>.value(
          value: authenticationRepository,
        ),
      ],
      child: MultiBlocProvider(
        providers: [
          BlocProvider<AuthenticationCubit<int>>.value(
            value: authenticationCubit,
          ),
          BlocProvider<SignUpCubit>(
            create: (_) => SignUpCubit(
              authenticationRepository: authenticationRepository,
              formData: Forms.getNormalData(),
            ),
          ),
          BlocProvider<SignInCubit>(
            create: (_) => SignInCubit(
              authenticationRepository: authenticationRepository,
            ),
          ),
        ],
        child: MaterialApp.router(
          title: 'Demo Authentication',
          debugShowCheckedModeBanner: false,
          routerDelegate: router.routerDelegate,
          routeInformationParser: router.routeInformationParser,
          routeInformationProvider: router.routeInformationProvider,
        ),
      ),
    );
  }
}

class GoRouterRefreshStream extends ChangeNotifier {
  GoRouterRefreshStream(Stream<dynamic> stream) {
    notifyListeners();
    _subscription = stream.asBroadcastStream().listen(
          (dynamic _) => notifyListeners(),
        );
  }

  late final StreamSubscription<dynamic> _subscription;

  @override
  void dispose() {
    _subscription.cancel();
    super.dispose();
  }
}
