// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:firebase_auth/firebase_auth.dart';
import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

class AuthenticationFirebaseDataSourceImpl
    extends AuthenticationRemoteDataSource {
  final FirebaseAuth _firebaseAuth;

  AuthenticationFirebaseDataSourceImpl({FirebaseAuth? firebaseAuth})
      : _firebaseAuth = firebaseAuth ?? FirebaseAuth.instance;

  Account _mapper(User user) => AccountModel(
        uid: user.uid,
        email: user.email,
      );

  @override
  Future<Account> signInWithEmailAndPassword({
    required String email,
    required String password,
  }) async {
    try {
      final userCredential = await _firebaseAuth.signInWithEmailAndPassword(
        email: email,
        password: password,
      );
      final user = userCredential.user;
      if (user.isNotNull) {
        return _mapper(user!);
      } else {
        throw Exception(); // Get caught just after.
      }
    } on FirebaseAuthException catch (e) {
      throw SignInWithEmailAndPasswordFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignInWithEmailAndPasswordFailureFirebase();
    }
  }

  @override
  Future<Account> signUp({
    required String email,
    required String password,
  }) async {
    try {
      final userCredential = await _firebaseAuth.createUserWithEmailAndPassword(
        email: email,
        password: password,
      );
      final user = userCredential.user;
      if (user.isNotNull) {
        return _mapper(user!);
      } else {
        throw Exception(); // Get caught just after.
      }
    } on FirebaseAuthException catch (e) {
      throw SignUpWithEmailAndPasswordFailureFirebase.fromCode(e.code);
    } catch (_) {
      throw SignUpWithEmailAndPasswordFailureFirebase();
    }
  }

  @override
  Future<void> signOut() async {
    try {
      await _firebaseAuth.signOut();
    } catch (_) {
      throw SignOutFailureFirebase();
    }
  }

  @override
  Future<String> getIdentityToken() async {
    try {
      final token = await _firebaseAuth.currentUser?.getIdToken();
      if (token.isNotNull) {
        return token!;
      } else {
        throw Exception(); // Get caught just after.
      }
    } catch (_) {
      // TODO(hpcl): implement a non ambiguous exception for this case
      throw ServerException();
    }
  }

  @override
  Stream<Account?> streamAccount() =>
      _firebaseAuth.userChanges().map<Account?>((user) {
        final Account? account = (user.isNotNull) ? _mapper(user!) : null;
        return account;
      });
}
