// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_authentication_bloc/src/data/models/account_wrapper_model.dart';
import 'package:wyatt_authentication_bloc/src/domain/data_sources/local/authentication_local_data_source.dart';
import 'package:wyatt_authentication_bloc/src/domain/data_sources/remote/authentication_remote_data_source.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/account.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/account_wrapper.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

class AuthenticationRepositoryImpl<T extends Object>
    extends AuthenticationRepository<T> {
  final AuthenticationLocalDataSource<T> _authenticationLocalDataSource;
  final AuthenticationRemoteDataSource _authenticationRemoteDataSource;

  final FutureResult<T?> Function(Account? account)? _onSignUpSuccess;
  final FutureResult<T?> Function(Account? account)? _onAccountChanges;

  AuthenticationRepositoryImpl(
    this._authenticationLocalDataSource,
    this._authenticationRemoteDataSource,
    this._onSignUpSuccess,
    this._onAccountChanges,
  );

  @override
  FutureResult<Account> signInWithEmailAndPassword({
    required String email,
    required String password,
  }) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account =
              await _authenticationRemoteDataSource.signInWithEmailAndPassword(
            email: email,
            password: password,
          );
          _authenticationLocalDataSource.storeAccount(account);
          return account;
        },
        (error) => error,
      );

  @override
  FutureResult<void> signOut() =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await _authenticationRemoteDataSource.signOut();
          _authenticationLocalDataSource.destroy();
        },
        (error) => error,
      );

  @override
  FutureResult<Account> signUp({
    required String email,
    required String password,
  }) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account = await _authenticationRemoteDataSource.signUp(
            email: email,
            password: password,
          );
          _authenticationLocalDataSource.storeAccount(account);
          if (_onSignUpSuccess.isNotNull) {
            final dataResult = await _onSignUpSuccess!.call(account);
            dataResult.fold(
              _authenticationLocalDataSource.storeData,
              (error) => throw error,
            );
          }
          return account;
        },
        (error) => error,
      );

  @override
  Result<void, AppException> destroyCache() =>
      Result.tryCatch<void, AppException, AppException>(
        _authenticationLocalDataSource.destroy,
        (error) => error,
      );

  @override
  Result<Account, AppException> getAccount() =>
      Result.tryCatch<Account, AppException, AppException>(
        _authenticationLocalDataSource.loadAccount,
        (error) => error,
      );

  @override
  Result<Account, AppException> setAccount(
    Account account,
  ) =>
      Result.tryCatch<Account, AppException, AppException>(
        () {
          _authenticationLocalDataSource.storeAccount(account);
          return account;
        },
        (error) => error,
      );

  @override
  Result<T, AppException> getData() =>
      Result.tryCatch<T, AppException, AppException>(
        _authenticationLocalDataSource.loadData,
        (error) => error,
      );

  @override
  Result<T, AppException> setData(
    T data,
  ) =>
      Result.tryCatch<T, AppException, AppException>(
        () {
          _authenticationLocalDataSource.storeData(data);
          return data;
        },
        (error) => error,
      );

  @override
  FutureResult<String> getIdentityToken() =>
      Result.tryCatchAsync<String, AppException, AppException>(
        _authenticationRemoteDataSource.getIdentityToken,
        (error) => error,
      );

  @override
  Stream<FutureResult<AccountWrapper<T>>> streamAccount() =>
      _authenticationRemoteDataSource.streamAccount().map((account) async {
        if (_onAccountChanges.isNotNull) {
          final dataResult = await _onAccountChanges!.call(account);
          return dataResult.map((data) => AccountWrapperModel(account, data));
        }
        return Ok<AccountWrapperModel<T>, AppException>(
          AccountWrapperModel<T>(account, null),
        );
      });
}
